@@ -20,11 +20,11 @@ Perl::Critic::Policy::ValuesAndExpressions::PreventSQLInjection - Prevent SQL in
 
 =head1 VERSION
 
-Version 1.2.0
+Version 1.3.0
 
 =cut
 
-our $VERSION = '1.2.0';
+our $VERSION = '1.3.0';
 
 
 =head1 AFFILIATION
@@ -51,10 +51,43 @@ But would leave alone:
 
 =head1 CONFIGURATION
 
-There is no configuration option available for this policy.
+=head2 quoting_methods
+
+A space-separated list of methods that are known to always return a safely
+quoted result.
+
+For example, to declare C<custom_quote()> as safe, add the following to your
+C<.perlcriticrc>:
+
+	[ValuesAndExpressions::PreventSQLInjection]
+	quoting_methods = 'custom_quote'
+
+By default, C<quote()> and C<quote_identifier> are considered safe, given their
+ubiquity in code that uses DBI. Note however that specifying manually a new
+list for C<quoting_methods> will override those defaults, so you will have to
+do this if you want to keep the two default methods but add your custom one to
+the list:
+
+	[ValuesAndExpressions::PreventSQLInjection]
+	quoting_methods = 'quote quote_identifier custom_quote'
 
 
-=head1 MARKING VARIABLES AS SAFE
+=head2 safe_functions
+
+A space-separated string listing the functions that always return a safely
+quoted value.
+
+For example, to declare C<quote_function()> and
+C<My::Package::quote_external_function()> as safe, add the following to your
+C<.perlcriticrc>:
+
+	[ValuesAndExpressions::PreventSQLInjection]
+	safe_functions = 'quote_function My::Package::quote_external_function'
+
+By default, no functions are considered safe.
+
+
+=head1 MARKING ELEMENTS AS SAFE
 
 You can disable this policy on a particular string with the usual PerlCritic
 syntax:
@@ -68,18 +101,57 @@ it was previously quoted with something such as:
 
 The risk there is that someone will later modify your code and introduce unsafe
 variables by accident, which will then not get reported. To prevent this, this
-module has a special C<## SQL safe ($var1, $var2, ...)> syntax which allows
-whitelisting specific variables:
+module has a special C<## SQL safe (...)> syntax described below.
+
+=head2 Marking variables as safe
+
+To indicate that a variable has been manually checked and determined to be
+safe, add a comment on the same line using this syntax: C<## SQL safe ($var1,
+$var2, ...)>.
+
+For example:
 
 	my $sql = "SELECT * FROM table WHERE field = $value"; ## SQL safe($value)
 
 That said, you should always convert your code to use placeholders instead
 where possible.
 
-Note: this policy supports both comma-separated and space-separated lists to
+=head2 Marking functions / class methods as safe
+
+To indicate that a function or class method has been manually checked and
+determined that it will always return a safe output, add comment on the same
+line using the C<## SQL safe(function_name>) syntax:
+
+	my $sql = "SELECT * FROM table WHERE field = "
+		. some_safe_method( $value ); ## SQL safe (&some_safe_method)
+
+	my $sql = "SELECT * FROM table WHERE field = "
+		. Package::Name::some_safe_method( $value ); ## SQL safe (&Package::Name::some_safe_method)
+
+Note that class methods (a function called with C<-E<gt>> on a package name)
+still need to be declared with C<::> in the list of safe elements:
+
+	my $sql = "SELECT * FROM table WHERE field = "
+		. Package::Name->some_safe_method( $value ); ## SQL safe (&Package::Name::some_safe_method)
+
+=head2 SQL safe syntax notes
+
+=over 4
+
+=item *
+
+This policy supports both comma-separated and space-separated lists to
 describe safe variables. In other words, C<## SQL safe ($var1, $var2, ...)> and
 C<## SQL safe ($var1 $var2 ...)> are strictly equivalent.
 
+=item *
+
+You can mix function names and variables in the comments to describe safe elements:
+
+	C<## SQL safe ($var1, &function_name, $var2, ...)>
+
+=back
+
 
 =head1 LIMITATIONS
 
@@ -98,14 +170,31 @@ This should probably be considered a violation:
 		$table
 	);
 
-=item * Detect use of $dbh->quote() and $dbh->quote_identifier()
+=item * Detect use of constants
 
-The following code will incorrectly trigger a violation:
+This should not be considered a violation, since constants cannot be modified
+by user input:
 
-	my $sql = 'SELECT * FROM ' . $dbh->quote_identifier( $test );
+	use Const::Fast;
+	const my $FOOBAR => 12;
 
-This is however quite a difficult problem, as the quote call can be hidden by
-several levels of indirection.
+	$dbh->do("SELECT name FROM categories WHERE id = $FOOBAR");
+
+=item * Detect SQL string modifications.
+
+Currently, this module only analyzes strings when they are declared, and does not account for later modifications.
+
+This should be reviewed as part of this module:
+
+	my $sql = "select from ";
+	$sql .= $table;
+
+As well as this:
+
+	my $sql = "select from ";
+	$sql = "$sql $table";
+
+=item
 
 =back
 
@@ -145,16 +234,17 @@ Readonly::Scalar my $VARIABLES_REGEX => qr/
 	)
 /x;
 
-# Name of the methods that make a variable safe to use in SQL strings.
-Readonly::Scalar my $QUOTING_METHODS_REGEX => qr/
-	^
-	(?:
-		quote_identifier
-		|
-		quote
-	)
-	$
-/x;
+# Default for the name of the methods that make a variable safe to use in SQL
+# strings.
+Readonly::Scalar my $QUOTING_METHODS_DEFAULT => q|
+	quote_identifier
+	quote
+|;
+
+# Default for the name of the packages and functions / class methods that are safe to
+# concatenate to SQL strings.
+Readonly::Scalar my $SAFE_FUNCTIONS_DEFAULT => q|
+|;
 
 # Regex to detect comments like ## SQL safe ($var1, $var2).
 Readonly::Scalar my $SQL_SAFE_COMMENTS_REGEX => qr/
@@ -184,7 +274,20 @@ Return an array with information about the parameters supported.
 
 sub supported_parameters
 {
-	return ();
+	return (
+		{
+			name            => 'quoting_methods',
+			description     => 'A space-separated string listing the methods that return a safely quoted value.',
+			default_string  => $QUOTING_METHODS_DEFAULT,
+			behavior        => 'string',
+		},
+		{
+			name            => 'safe_functions',
+			description     => 'A space-separated string listing the functions that return a safely quoted value',
+			default_string  => $SAFE_FUNCTIONS_DEFAULT,
+			behavior        => 'string',
+		},
+	);
 }
 
 
@@ -248,6 +351,8 @@ sub violates
 {
 	my ( $self, $element, $doc ) = @_;
 
+	parse_config_parameters( $self );
+
 	parse_comments( $self, $doc );
 
 	# Make sure the first string looks like a SQL statement before investigating
@@ -255,49 +360,8 @@ sub violates
 	return ()
 		if !is_sql_statement( $element );
 
-	my $sql_injections = [];
-	my $token = $element;
-	while ( defined( $token ) && $token ne '' )
-	{
-		# If the token is a string, we need to analyze it for interpolated
-		# variables.
-		if ( $token->isa( 'PPI::Token::HereDoc' ) || $token->isa( 'PPI::Token::Quote' ) ) ## no critic (ControlStructures::ProhibitCascadingIfElse)
-		{
-			push( @$sql_injections, @{ analyze_sql_injections( $self, $token ) || [] } );
-		}
-		# If it is a concatenation operator, continue to the next token.
-		elsif ( $token->isa('PPI::Token::Operator') && $token->content() eq '.' )
-		{
-			# Skip to the next token.
-		}
-		# If it is a semicolon, we're at the end of the statement and we can finish
-		# the process.
-		elsif ( $token->isa('PPI::Token::Structure') && $token->content() eq ';' )
-		{
-			last;
-		}
-		# If it is a list-separating comma, this list element ends here and we can
-		# finish the process.
-		elsif ( $token->isa('PPI::Token::Operator') && $token->content() eq ',' )
-		{
-			last;
-		}
-		# If it is a symbol, it is concatenated to a SQL statement which is an
-		# injection risk.
-		elsif ( $token->isa('PPI::Token::Symbol') )
-		{
-			my ( $variable, $is_quoted ) = get_complete_variable( $token );
-			if ( !$is_quoted )
-			{
-				my $safe_variables = get_safe_variables( $self, $token->line_number() );
-				push( @$sql_injections, $variable )
-					if !exists( $safe_variables->{ $variable } );
-			}
-		}
-
-		# Move to examining the next sibling token.
-		$token = $token->snext_sibling();
-	}
+	# Find SQL injection vulnerabilities.
+	my $sql_injections = detect_sql_injections( $self, $element );
 
 	# Return violations if any.
 	return defined( $sql_injections ) && scalar( @$sql_injections ) != 0
@@ -313,13 +377,172 @@ sub violates
 }
 
 
+=head1 INTERNAL FUNCTIONS
+
+=head2 detect_sql_injections()
+
+Detect SQL injections vulnerabilities tied to the PPI element specified.
+
+	my $sql_injections = detect_sql_injections( $policy, $element );
+
+=cut
+
+sub detect_sql_injections
+{
+	my ( $self, $element ) = @_;
+
+	my $sql_injections = [];
+	my $token = $element;
+	while ( defined( $token ) && $token ne '' )
+	{
+		# If the token is a string, we need to analyze it for interpolated
+		# variables.
+		if ( $token->isa( 'PPI::Token::HereDoc' ) || $token->isa( 'PPI::Token::Quote' ) ) ## no critic (ControlStructures::ProhibitCascadingIfElse)
+		{
+			push( @$sql_injections, @{ analyze_string_injections( $self, $token ) || [] } );
+		}
+		# If it is a concatenation operator, continue to the next token.
+		elsif ( $token->isa('PPI::Token::Operator') && $token->content() eq '.' )
+		{
+			# Skip to the next token.
+		}
+		# If it is a semicolon, we're at the end of the statement and we can finish
+		# the process.
+		elsif ( $token->isa('PPI::Token::Structure') && $token->content() eq ';' )
+		{
+			last;
+		}
+		# If we detect a ':' operator, we're at the end of the second argument in a
+		# ternary "... ? ... : ..." and we need to finish the process here as the
+		# third argument is not concatenated to the this string and will be
+		# analyzed separately.
+		elsif ( $token->isa('PPI::Token::Operator') && $token->content() eq ':' )
+		{
+			last;
+		}
+		# If it is a list-separating comma, this list element ends here and we can
+		# finish the process.
+		elsif ( $token->isa('PPI::Token::Operator') && $token->content() eq ',' )
+		{
+			last;
+		}
+		# If it is a symbol, it is concatenated to a SQL statement which is an
+		# injection risk.
+		elsif ( $token->isa('PPI::Token::Symbol') )
+		{
+			my ( $variable, $is_quoted ) = get_complete_variable( $self, $token );
+			if ( !$is_quoted )
+			{
+				my $safe_elements = get_safe_elements( $self, $token->line_number() );
+				push( @$sql_injections, $variable )
+					if !exists( $safe_elements->{ $variable } );
+			}
+		}
+		# If it is a word, it may be a function/method call on a package, which is
+		# an injection risk.
+		elsif ( $token->isa('PPI::Token::Word') )
+		{
+			# Find out if the PPO::Token::Word is the beginning of a call or not.
+			my ( $function_name, $is_quoted ) = get_function_name( $self, $token );
+			if ( defined( $function_name ) && !$is_quoted )
+			{
+				my $safe_elements = get_safe_elements( $self, $token->line_number() );
+				push( @$sql_injections, $function_name )
+					if !exists( $safe_elements->{ '&' . $function_name } );
+			}
+		}
+
+		# Move to examining the next sibling token.
+		$token = $token->snext_sibling();
+	}
+
+	return $sql_injections;
+}
+
+
+=head2 get_function_name()
+
+Retrieve full name (including the package name) of a class function/method
+based on a PPI::Token::Word object, and indicate if it is a call that returns
+quoted data making it safe to include directly into SQL strings.
+
+	my ( $function_name, $is_quoted ) = get_function_name( $policy, $token );
+
+=cut
+
+sub get_function_name
+{
+	my ( $self, $token ) = @_;
+
+	croak 'The first parameter needs to be a PPI::Token::Word object'
+		if !$token->isa('PPI::Token::Word');
+
+	my $next_sibling = $token->snext_sibling();
+	return ()
+		if !defined( $next_sibling ) || ( $next_sibling eq '' );
+
+	my ( $package, $function_name );
+
+	# Catch Package::Name->method().
+	if ( $next_sibling->isa('PPI::Token::Operator') && ( $next_sibling->content() eq '->' ) )
+	{
+		my $function = $next_sibling->snext_sibling();
+
+		return ()
+			if !defined( $function ) || ( $function eq '' );
+		return ()
+			if !$function->isa('PPI::Token::Word');
+
+		$package = $token->content();
+		$function_name = $function->content();
+
+		$function->{'_handled'} = 1;
+	}
+	# Catch Package::Name::function().
+	elsif ( $next_sibling->isa('PPI::Structure::List') )
+	{
+		# Package::Name->method() will result in two PPI::Token::Word being
+		# detected, one for 'Package::Name' and one for 'method'. 'Package::Name'
+		# will be caught in the if() block above, but 'method' would get caught
+		# separately by this block. To prevent this, we scan the previous sibling
+		# here and skip if we find that it is a '->' operator.
+		my $previous_sibling = $token->sprevious_sibling();
+		return ()
+			if $previous_sibling->isa('PPI::Token::Operator') && ( $previous_sibling->content() eq '->' );
+
+		my $content = $token->content();
+
+		# Catch function calls in the same namespace.
+		if ( $content !~ /::/ )
+		{
+			( $package, $function_name ) = ( undef, $content );
+		}
+		# Catch function calls in a different namespace.
+		else
+		{
+			( $package, $function_name ) = $content =~ /^(.*)::([^:]+)$/;
+		}
+	}
+	else
+	{
+		return ();
+	}
+
+	my $full_name = join( '::', grep { defined( $_ ) } ( $package, $function_name ) );
+	my $is_safe = defined( $self->{'_safe_functions_regex'} ) && ( $full_name =~ $self->{'_safe_functions_regex'} )
+		? 1
+		: 0;
+	return ( $full_name, $is_safe );
+}
+
+
 =head2 get_complete_variable()
 
 Retrieve a complete variable starting with a PPI::Token::Symbol object, and
 indicate if the variable has used a quoting method to make it safe to use
 directly in SQL strings.
 
-	my ( $variable, $is_quoted ) = get_complete_variable( $token );
+	my ( $variable, $is_quoted ) = get_complete_variable( $policy, $token );
 
 For example, if you have $variable->{test}->[0] in your code, PPI will identify
 $variable as a PPI::Token::Symbol, and calling this function on that token will
@@ -329,7 +552,7 @@ return the whole "$variable->{test}->[0]" string.
 
 sub get_complete_variable
 {
-	my ( $token ) = @_;
+	my ( $self, $token ) = @_;
 
 	croak 'The first parameter needs to be a PPI::Token::Symbol object'
 		if !$token->isa('PPI::Token::Symbol');
@@ -352,7 +575,8 @@ sub get_complete_variable
 		}
 		elsif ( $sibling->isa('PPI::Token::Word')
 			&& $sibling->method_call()
-			&& ( $sibling->content =~ $QUOTING_METHODS_REGEX )
+			&& defined( $self->{'_quoting_methods_regex'} )
+			&& ( $sibling->content =~ $self->{'_quoting_methods_regex'} )
 		)
 		{
 			$is_quoted = 1;
@@ -404,7 +628,6 @@ sub get_token_content
 	if ( $token->isa('PPI::Token::HereDoc') )
 	{
 		my @heredoc = $token->heredoc();
-		pop( @heredoc ); # Remove the heredoc termination tag.
 		$content = join( '', @heredoc );
 	}
 	elsif ( $token->isa('PPI::Token::Quote' ) )
@@ -421,19 +644,19 @@ sub get_token_content
 }
 
 
-=head2 analyze_sql_injections()
+=head2 analyze_string_injections()
 
-Analyze a token and returns an arrayref of variables that are potential SQL
-injection vectors.
+Analyze a token representing a string and returns an arrayref of variables that
+are potential SQL injection vectors.
 
-	my $sql_injection_vector_names = analyze_sql_injections(
-		$policy, # this object
+	my $sql_injection_vector_names = analyze_string_injections(
+		$policy,
 		$token,
 	);
 
 =cut
 
-sub analyze_sql_injections
+sub analyze_string_injections
 {
 	my ( $policy, $token ) = @_;
 
@@ -457,7 +680,7 @@ sub analyze_sql_injections
 		#       determine the ending line number instead of the beginning line
 		#       number.
 		my $extra_height_span =()= $content =~ /\n/g;
-		my $safe_variables = get_safe_variables(
+		my $safe_elements = get_safe_elements(
 			$policy, #$self
 			$token->line_number()
 				# Heredoc comments will be on the same line as the opening marker.
@@ -466,7 +689,7 @@ sub analyze_sql_injections
 
 		# Find all the variables that appear in the string.
 		my $unsafe_variables = [
-			grep { !$safe_variables->{ $_ } }
+			grep { !$safe_elements->{ $_ } }
 			@{ extract_variables( $content ) }
 		];
 
@@ -541,18 +764,18 @@ sub extract_variables
 }
 
 
-=head2 get_safe_variables()
+=head2 get_safe_elements()
 
-Return a hashref with safe variable names as the keys.
+Return a hashref with safe element names as the keys.
 
-	my $safe_variables = get_safe_variables(
-		$self,
+	my $safe_elements = get_safe_elements(
+		$policy,
 		$line_number,
 	);
 
 =cut
 
-sub get_safe_variables
+sub get_safe_elements
 {
 	my ( $self, $line_number ) = @_;
 
@@ -566,7 +789,7 @@ sub get_safe_variables
 	return {}
 		if !exists( $self->{'_sqlsafe'}->{ $line_number } );
 
-	# Return a hash of safe variable names.
+	# Return a hash of safe element names.
 	return {
 		map
 			{ $_ => 1 }
@@ -577,11 +800,11 @@ sub get_safe_variables
 
 =head2 parse_comments()
 
-Parse the comments for the current document and identify variables marked as
+Parse the comments for the current document and identify elements marked as
 SQL safe.
 
 	parse_comments(
-		$self,
+		$policy,
 		$ppi_document,
 	);
 
@@ -601,17 +824,60 @@ sub parse_comments
 	foreach my $comment ( @$comments )
 	{
 		# Determine if the line is a "SQL safe" comment.
-		my ( $safe_variables ) = $comment =~ $SQL_SAFE_COMMENTS_REGEX;
-		next if !defined( $safe_variables );
+		my ( $safe_elements ) = $comment =~ $SQL_SAFE_COMMENTS_REGEX;
+		next if !defined( $safe_elements );
 
-		# Store list of safe variables for that line.
+		# Store list of safe elements for that line.
 		push(
 			@{ $self->{'_sqlsafe'}->{ $comment->line_number() } },
-			split( /[\s,]+(?=[\$\@\%])/, $safe_variables )
+			split( /[\s,]+(?=[\$\@\%\&])/, $safe_elements )
 		);
 	}
 
-	#print STDERR "SQL safe variables: ", Dumper( $self->{'_sqlsafe'} ), "\n";
+	#print STDERR "SQL safe elements: ", Dumper( $self->{'_sqlsafe'} ), "\n";
+	return;
+}
+
+
+=head2 parse_config_parameters()
+
+Parse the parameters from the C<.perlcriticrc> file, if any are specified
+there.
+
+	parse_config_parameters( $policy );
+
+=cut
+
+sub parse_config_parameters
+{
+	my ( $self ) = @_;
+
+	if ( !exists( $self->{'_quoting_methods_regex'} ) )
+	{
+		if ( $self->{'_quoting_methods'} =~ /\w/ )
+		{
+			my $regex_components = join( '|', grep { $_ =~ /\w/ } split( /,?\s+/, $self->{'_quoting_methods'} ) );
+			$self->{'_quoting_methods_regex'} = qr/^(?:$regex_components)$/x;
+		}
+		else
+		{
+			$self->{'_quoting_methods_regex'} = undef;
+		}
+	}
+
+	if ( !exists( $self->{'_safe_functions_regex'} ) )
+	{
+		if ( $self->{'_safe_functions'} =~ /\w/ )
+		{
+			my $regex_components = join( '|', grep { $_ =~ /\w/ } split( /,?\s+/, $self->{'_safe_functions'} ) );
+			$self->{'_safe_functions_regex'} = qr/^(?:$regex_components)$/x;
+		}
+		else
+		{
+			$self->{'_safe_functions_regex'} = undef;
+		}
+	}
+
 	return;
 }
 
@@ -662,7 +928,7 @@ C<< <aubertg at cpan.org> >>.
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2013 Guillaume Aubert.
+Copyright 2013-2014 Guillaume Aubert.
 
 This program is free software: you can redistribute it and/or modify it under
 the terms of the GNU General Public License version 3 as published by the Free
